%ESTIM_Z   Estimate PEA deformation from an interference pattern.
%   Z = ESTIM_Z(T,V,X,OPTS) estimates the PEA deformation Z starting from
%   an interference signal X and the PEA applied voltage V.
%   T is the vector of time instants (of the same length as V and X and
%   OPTS is a structure with the following fields:
%    - lambda: the laser wavelength;
%    - window: the time windows used to search extrema in the interference
%              signal X;
%    - en_plot: if 1, plots are enabled;
%    - n: order of the low-pass filter used to filter the interference signal;
%    - wn: normalized cut frequency for the filter. See documentation of
%          funcrtion "fir" for details;
%    Voltage V must be one period of a triangular wave composed of a
%    descending branch followed by an ascending branch.
%
%    If OPTS.en_plot = 1, 6 figures are generated, useful to monitor the
%    behaviour of the function:
%     - Figure 1: descending and ascending branches of V, and
%                 corresponding interference signal;
%     - Figure 2: filtered interference signal, If the filtering is not
%                 adequate, modify OPTS.n and/or OPTS.wn;
%     - Figure 3: filtered signal X with identified extrema (red crosses).
%                 If some extrema are missing or there are too many
%                 extrema, modify OPTS.window and/or the filtering
%                 parameters;
%     - Figure 4: polynomial fitting on ascending and descending branches
%                 of signal Z. If the fit is not accurate, modify the
%                 fitting function (line 158 of this function);
%     - Figure 5: the same of Fig. 4, shown as a function of V instead of T;
%     - Figure 6: final hysteresis loop Z-V

%   Acknowledgements:
%   Alberto Oliveri, University of Genoa (Italy), 2018.
%   alberto.oliveri@unige.it

function z = estim_z(t,v,x,opts)

close all

% Laser wavelength
lambda = opts.lambda;

% Window for finding extrema of x
window = opts.window;

% Enable plot
en_plot = opts.en_plot;

% Filter order and normalized frequency
n = opts.n;
wn = opts.wn;

% Displacement resolution
dz = lambda/4;

% Find index of minimum value of v
[vmin, idx] = min(v);

% Split data (descending / ascending)
tdown = t(1:idx);
vdown = v(1:idx);
xdown = x(1:idx);
tup = t(idx+1:end);
vup = v(idx+1:end);
xup = x(idx+1:end);

if en_plot
    figure
    subplot(2,2,1)
    plot(tup,vup)
    xlabel('t')
    ylabel('v')
    title('Ascending')
    subplot(2,2,2)
    plot(tdown,vdown)
    xlabel('t')
    ylabel('v')
    title('Descending')
    subplot(2,2,3)
    plot(tup,xup)
    xlabel('t')
    ylabel('x')
    subplot(2,2,4)
    plot(tdown,xdown)
    xlabel('t')
    ylabel('x')
end

% Filter interference data

% Filter ascending and descending x
b = fir1(n,wn);
xxup = filter(b,1,xup,-0*ones(numel(b)-1,1));
xxup = [xxup(n/2:end); xxup(end)*ones(n/2-1,1)];

xxdown = filter(b,1,xdown,-0*ones(numel(b)-1,1));
xxdown = [xxdown(n/2:end); xxdown(end)*ones(n/2-1,1)];

if en_plot
    figure
    subplot(1,2,1)
    hold on
    plot(tup,xup)
    plot(tup,xxup)
    xlabel('t')
    ylabel('x')
    title('Ascending')
    legend('original','filtered')
    subplot(1,2,2)
    hold on
    plot(tdown,xdown)
    plot(tdown,xxdown)
    xlabel('t')
    ylabel('x')
    title('Descending')
    legend('original','filtered')
end

% Find extrema of x

% Find extrema
[pksdown, locsdown] = findmaxmin(xxdown,1,window);
[locsdown,idx] = unique(locsdown);
pksdown = pksdown(idx);
locsdown(end) = [];
pksdown(end) = [];

% Find extrema
[pksup, locsup] = findmaxmin(xxup,1,window);
[locsup,idx] = unique(locsup);
pksup = pksup(idx);
locsup(end) = [];
pksup(end) = [];

if en_plot
    figure
    subplot(1,2,1)
    hold on
    plot(tup,xxup)
    plot(tup(locsup),pksup,'rx')
    xlabel('t')
    ylabel('x')
    title('Ascending')
    subplot(1,2,2)
    hold on
    plot(tdown,xxdown)
    plot(tdown(locsdown),pksdown,'rx')
    xlabel('t')
    ylabel('x')
    title('Descending')
end

% Interpolate branches of z with a second order polynomial

idx = 1:numel(locsdown);
zdown = -dz*(idx-1);

fitobject = fit(tdown(locsdown(:)),zdown(:),'poly2');
zfitdown = feval(fitobject,tdown);

idx = 1:numel(locsup);
zup = dz*(idx-1);

fitobject = fit(tup(locsup(:)),zup(:),'poly2');
zfitup = feval(fitobject,tup);

if en_plot
    figure
    hold on
    plot(tup(locsup),zup,'rx')
    plot(tup,zfitup,'r')
    plot(tdown(locsdown),zdown,'gx')
    plot(tdown,zfitdown,'g')
    xlabel('t')
    ylabel('z')
    title('Polynomial fitting')
end

% Join branches

o1 = zfitdown(1)-zfitup(end);
o2 = zfitdown(end)-zfitup(1);
offset = 0.5*(o1+o2);
zfitup = zfitup+offset;
zup = zup+offset;

if en_plot
    figure
    hold on
    plot(vup(locsup),zup,'xr')
    plot(vup,zfitup,'r')
    plot(vdown(locsdown),zdown,'xg')
    plot(vdown,zfitdown,'g')
    xlabel('v')
    ylabel('x')
    title('Polynomial fitting - Hysteresis loop')
end

% Join data

zfit = [zfitdown;zfitup];

zmean = 0.5*(min(zfit)+max(zfit));
z = zfit-zmean;

if en_plot
    figure
    hold on
    plot(v,z)
    xlabel('v')
    ylabel('z')
    title('Hysteresis loop')
end

    function [pks,loc] = findmaxmin(x,wstart,windowsize)
        
        windowsizeorig = windowsize;
        
        % Look if first extremum is a minimum or a maximum
        xwin = x(wstart:wstart+windowsize-1);
        
        [val,idxM] = max(xwin);
        [val,idxm] = min(xwin);
        
        if idxm < round(windowsize/2)
            idxm = 1e6;
        end
        if idxM < round(windowsize/2)
            idxM = 1e6;
        end
        
        if idxM > idxm
            curextr = 'min';
        else
            curextr = 'max';
        end
        
        go = 1;
        nextgo = 0;
        
        loc = [];
        pks = [];
        
        while go
            
            found = 0;
            
            idxtmpprev = 0;
            windowsize = windowsizeorig;
            
            while not(found)
                
                xwin = x(wstart:wstart+windowsize-1);
                
                if strcmp(curextr,'min')
                    [valtmp,idxtmp] = min(xwin);
                else
                    [valtmp,idxtmp] = max(xwin);
                end
                
                if idxtmp == idxtmpprev
                    found = 1;
                else
                    if nextgo
                        break
                    else
                        windowsize = windowsize+windowsizeorig;
                        if wstart+windowsize-1 >= numel(x)
                            windowsize = numel(x)+1-wstart;
                            nextgo = 1;
                        end
                        idxtmpprev = idxtmp;
                    end
                end
                
                
            end
            
            if strcmp(curextr,'min')
                curextr = 'max';
            else
                curextr = 'min';
            end
            
            if found
                
                pks = [pks; valtmp];
                loc = [loc;idxtmp+wstart-1];
                
            end
            
            wstart = loc(end);
            
            if nextgo
                go = 0;
            end
            
        end
    end
end